/*
 *	Copyright (c) 1993-1996 by Research Computer Center (Expert Team)
 *	Moscow State University, Moscow, Russia
 *	All rights reserved.
 *
 *	%W%     %E%     (c) SRCC
 */

#define STRINGS_INLINE

#include <assert.h>
#include <string.h>
#include <typeinfo>

#ifdef DEBUG
#include <stdio.h>
#endif

_NS_STD_BEGIN

struct rtti_info {
	type_info *t_info;
	int offset;
};

extern "C" int
_throw_bad_cast ()
{
	throw bad_cast();
	return 0;
}

/*
 * Meaning of mode parameter:
 *	0 meant that classes are not bound with inheritance. We shall check
 *		only public base classes.
 *	-1 means that we cast virtual base class to the derived class. In
 *		case we shall check all base classes (not only public) and
 *		check that class_to is unique base class.
 *	n (n>0) means that we cast from base class to derived class. We shall
 *		not check unique subobject, and return always n-1 (static
 *		subclass offset.
 */

extern "C" int
_dynamic_cast (type_info *ptr0, char **name_from0, char **name_to0, int mode)
{
	int state = 0;
	int ret = -1;
        struct rtti_info *ptr = (rtti_info*)(ptr0+1);
	char *name_from = *name_from0;
	char *name_to = *name_to0;

#ifdef DEBUG
	printf ("name_from = %s; name_to = %s;\n", name_from, name_to);
	printf ("real class name: %s; mode = %d\n", ptr0->name(), mode);
#endif

	if (ptr0->name() == name_from || strcmp(ptr0->name(), name_from)==0)
		ret = 0;
	else while (ptr->t_info != 0 || ptr->offset != 0) {

		if (ptr->t_info == 0) {
			if (mode == 0) {

#ifdef DEBUG
				printf ("Delimeter skiped.\n");
#endif

				/* skip the public/nonpublic bases delimeter */
				ptr++;
				continue;
			} else
				break;
		};

#ifdef DEBUG
		printf ("ptr = %p; ptr->name = %p; ptr->offset = %d;\n",
				ptr, ptr->t_info->name(), ptr->offset);
		printf ("1. ptr->name = %s;\n", ptr->t_info->name());
#endif

		if (ptr->t_info->name() == name_from ||
		    strcmp(ptr->t_info->name(), name_from) == 0)
			ret = ptr->offset;

		ptr++;
	};

	if (ret==-1 && mode > 0)
		return -1;

		// Added by AlCh 05.06.98.
		// Check downcast for most derived object type.
	if ( ptr0->name() == name_to || strcmp(ptr0->name(), name_to)==0 ) {
		if ( mode > 0 ) {
			return mode-1;
		}

		ret = 0 - ret;
		return ret;
	}

	ptr = (rtti_info*)(ptr0+1);

	while (ptr->t_info != 0) {
		/* look only for public base class subobjects */

		if (ptr->t_info->name() == name_to ||
		    strcmp(ptr->t_info->name(), name_to) == 0) {

#ifdef DEBUG
			printf ("2. ptr->name = %s;\n", ptr->t_info->name());
#endif

			if (mode > 0)
				/* there is a derived class subobject */
				return mode-1;

			ret = ptr->offset-ret;

			if (state == 1)
				return -1;
			else
				state = 1;
		};
		ptr++;
	};

	if (state == 1) {
#ifdef DEBUG
		printf ("returning %d;\n", ret);
#endif
		return ret;
	} else {
#ifdef DEBUG
		printf ("returning -1;\n", ret);
#endif
		return -1;
	}
}

extern "C" int
_rtti_find (struct rtti_info *ptr, char *name)
{
	int ret;
	int state = 0;

	/* wall trough all types */
	while (ptr->t_info->name() != 0 || ptr->offset != 0) {
		if (ptr->t_info->name() == name ||
		    strcmp(ptr->t_info->name(), name) == 0) {
			ret = ptr->offset;

			if (state == 1)
				return -1;
			else
				state = 1;
		};

		ptr++;
	};

	/* assume there is no sub-object at offset -1 */
	if (state == 1)
		return ret;

  return -1;
}

_NS_STD_END
