/*----------------------------------------------------------------
 *
 * NeuroChip NM6403, C run-time library
 * 1997-1998 (c) RC Module Inc.
 * malloc.h,
 *
 *---------------------------------------------------------------- */

#ifndef _MALLOC_H
#define _MALLOC_H


#define	__P(args)	args
typedef  void *     __ptr_t;

#include <stddef.h>
#define	__malloc_size_t	size_t

#ifndef	NULL
#define	NULL	0
#endif


#ifdef	__cplusplus
extern "C"
{
#endif

/* Allocate SIZE bytes of memory.  */
extern __ptr_t malloc __P ((size_t __size));
extern __ptr_t malloc0 __P ((size_t __size));
extern __ptr_t malloc1 __P ((size_t __size));
extern __ptr_t malloc2 __P ((size_t __size));
extern __ptr_t malloc3 __P ((size_t __size));

/* Allocate memory for num elements of __size bytes each and fill it by zeroes */
extern __ptr_t calloc __P ((size_t __size,size_t num));
extern __ptr_t calloc0 __P ((size_t __size,size_t num));
extern __ptr_t calloc1 __P ((size_t __size,size_t num));
extern __ptr_t calloc2 __P ((size_t __size,size_t num));
extern __ptr_t calloc3 __P ((size_t __size,size_t num));

/* Re-allocate the previously allocated block
   in __ptr_t, making the new block SIZE bytes long.  */
extern __ptr_t realloc __P ((__ptr_t __ptr, size_t __size));
extern __ptr_t realloc0 __P ((__ptr_t __ptr, size_t __size));
extern __ptr_t realloc1 __P ((__ptr_t __ptr, size_t __size));
extern __ptr_t realloc2 __P ((__ptr_t __ptr, size_t __size));
extern __ptr_t realloc3 __P ((__ptr_t __ptr, size_t __size));

/* Free a block allocated by `malloc', `realloc' or `calloc'.  */
extern void free __P ((__ptr_t __ptr));
extern void free0 __P ((__ptr_t __ptr));
extern void free1 __P ((__ptr_t __ptr));
extern void free2 __P ((__ptr_t __ptr));
extern void free3 __P ((__ptr_t __ptr));

#ifdef _MALLOC_INTERNAL

    extern int _HEAP_SIZE, _HEAP_SIZE1, _HEAP_SIZE2, _HEAP_SIZE3;

#ifdef __NO_NEURO          /* WIN32 assumed */
    extern int _heap[], _heap1[], _heap2[], _heap3[];

    #define _D_HEAP_BASE0	((char *)(unsigned int)_heap)
    #define _D_HEAP_SIZE0	((size_t)(_HEAP_SIZE))
    #define _D_HEAP_BASE1 	((char *)(unsigned int)_heap1)
    #define _D_HEAP_SIZE1   ((size_t)(_HEAP_SIZE1))
    #define _D_HEAP_BASE2 	((char *)(unsigned int)_heap2)
    #define _D_HEAP_SIZE2   ((size_t)(_HEAP_SIZE2))
    #define _D_HEAP_BASE3 	((char *)(unsigned int)_heap3)
    #define _D_HEAP_SIZE3   ((size_t)(_HEAP_SIZE3))

#else
    extern int  _heap, _heap1, _heap2, _heap3;

    #define _D_HEAP_BASE0	((char *)(unsigned int)&_heap)
    #define _D_HEAP_SIZE0	((size_t)(&_HEAP_SIZE))
    #define _D_HEAP_BASE1 	((char *)(unsigned int)&_heap1)
    #define _D_HEAP_SIZE1	((size_t)(&_HEAP_SIZE1))
    #define _D_HEAP_BASE2 	((char *)(unsigned int)&_heap2)
    #define _D_HEAP_SIZE2	((size_t)(&_HEAP_SIZE2))
    #define _D_HEAP_BASE3 	((char *)(unsigned int)&_heap3)
    #define _D_HEAP_SIZE3	((size_t)(&_HEAP_SIZE3))

#endif

/* returns 1 if pointer ptr points somewhere in the 0 heap */
#define __IN_HEAP0(ptr)     \
         ( ( (ptr>=_D_HEAP_BASE0) && (ptr<_D_HEAP_BASE0+_D_HEAP_SIZE0) ) ? 1:0 )
/* returns 1 if pointer ptr points somewhere in the first heap */
#define __IN_HEAP1(ptr)     \
         ( ( (ptr>=_D_HEAP_BASE1) && (ptr<_D_HEAP_BASE1+_D_HEAP_SIZE1) ) ? 1:0 )
/* returns 1 if pointer ptr points somewhere in the 2 heap */
#define __IN_HEAP2(ptr)     \
         ( ( (ptr>=_D_HEAP_BASE2) && (ptr<_D_HEAP_BASE2+_D_HEAP_SIZE2) ) ? 1:0 )
/* returns 1 if pointer ptr points somewhere in the 3 heap */
#define __IN_HEAP3(ptr)     \
         ( ( (ptr>=_D_HEAP_BASE3) && (ptr<_D_HEAP_BASE3+_D_HEAP_SIZE3) ) ? 1:0 )

/* The allocator divides the heap into blocks of fixed size; large
   requests receive one or more whole blocks, and small requests
   receive a fragment of a block.  Fragment sizes are powers of two,
   and all fragments of a block are the same size.  When all the
   fragments in a block have been freed, the block itself is freed.  */

#define BLOCKLOG	7
#define BLOCKSIZE	(1 << BLOCKLOG)
#define BLOCKIFY(SIZE)	(((SIZE) + BLOCKSIZE - 1) / BLOCKSIZE)

/* Determine the amount of memory spanned by the initial heap table
   (not an absolute limit).  */

#define HEAP            16384

/* Number of contiguous free blocks allowed to build up at the end of
   memory before they will be returned to the system.  */

#define FINAL_FREE_BLOCKS    8

typedef struct
      {
	__malloc_size_t size;	/* Size (in blocks) of a free cluster.  */
	__malloc_size_t next;	/* Index of next free cluster.  */
	__malloc_size_t prev;	/* Index of previous free cluster.  */
      }StructFree;

/* Data structure giving per-block information.  */
typedef union
  {
    /* Heap information for a busy block.  */
    struct
      {
	/* Zero for a large block, or positive giving the
	   logarithm to the base two of the fragment size.  */
	int type;
	union
	  {
	    struct
	      {
		__malloc_size_t nfree; /* Free frags in a fragmented block.  */
		__malloc_size_t first; /* First free fragment of the block.  */
	      } frag;
	    /* Size (in blocks) of a large cluster.  */
	    __malloc_size_t size;
	  } info;
      } busy;
    /* Heap information for a free block
       (that may be the first of a free cluster).  */
    StructFree free;
  } malloc_info;

/* Doubly linked lists of free fragments.  */
struct list
  {
    struct list *next;
    struct list *prev;
  };

typedef struct  {
        int                 __malloc_initialized;
        /* Underlying allocation function; successive calls should
           return contiguous pieces of memory.  */
        __ptr_t             (*__morecore) __P ((ptrdiff_t __size));
        /* Pointer to first block of the heap.  */
        char                *_malloc_heapbase;
        /* Table indexed by block number giving per-block information.  */
        malloc_info         *_heapinfo;
        /* Current search index for the heap table.  */
        __malloc_size_t     _heapindex;
        /* Limit of valid info table indices.  */
        __malloc_size_t     _heaplimit;
        /* Instrumentation.  */
        __malloc_size_t     _chunks_used;
        __malloc_size_t     _bytes_used;
        __malloc_size_t     _chunks_free;
        __malloc_size_t     _bytes_free;
        /* Free list headers for each fragment size.  */
        struct list         _fraghead[BLOCKLOG];

        /* Number of blocks */
        unsigned int        heapsize;
} _heap_control_block;

extern _heap_control_block _hcb0;       /* 0 heap */
extern _heap_control_block _hcb1;       /* local heap */
extern _heap_control_block _hcb2;       /* global heap */
extern _heap_control_block _hcb3;       /* third heap */

/* Address to block number and vice versa.  */
#define BLOCK(A,cblock) \
        (((char *) (A) - cblock->_malloc_heapbase) / BLOCKSIZE + 1)
#define ADDRESS(B,cblock) \
        ((__ptr_t) (((B) - 1) * BLOCKSIZE + cblock->_malloc_heapbase))

/* Internal version of `free' used in `morecore' (malloc.c). */
extern void __free_internal (__ptr_t __ptr,_heap_control_block *cb);
extern __ptr_t __malloc_internal __P ((__malloc_size_t,_heap_control_block*));

/* Default value of `__morecore'.  */
extern __ptr_t __default_morecore0 __P ((ptrdiff_t __size));
extern __ptr_t __default_morecore1 __P ((ptrdiff_t __size));
extern __ptr_t __default_morecore2 __P ((ptrdiff_t __size));
extern __ptr_t __default_morecore3 __P ((ptrdiff_t __size));

#endif /* _MALLOC_INTERNAL.  */

#ifdef	__cplusplus
}
#endif

#endif  // _MALLOC_H
