//////////////////////////////////////////////////////////////////////////
//                                                                      //
// tanhf.cpp -                                                          //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:10p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

const unsigned int TLN2 = 0x2c5c85fe;   //  Ln(2) * 0x40000000

//     
//  _TLN[i] = Ln(1+(0.5)^i)
static int _TLN[29] ={0x19F323EC,	0xE47FBE3,	0x789C1DB,	0x3E14618,
		0x1F829B0,	0xFE0545,	0x7F80A9,	0x3FE015,
		0x1FF802,	0xFFE00,	0x7FF80,	0x3FFE0,
		0x1FFF8,	0xFFFE,	0x7FFF,	0x3FFF,	0x1FFF,	0xFFF,	0x7FF,
		0x3FF,	0x1FF,	0xFF,	0x7F,	0x3F,	0x1F,	0xF,	0x7,
		0x3,	0x1	};

//      
static int _MathLib_iK[13]    = {  0x2C9E15D2,	0x227E7B3D,	0x2097C51C,	0x20257BCF,
                        	0x200957BA,	0x2002557B,	0x20009557,	0x20002555,
                            0x20000955,	0x20000255,	0x20000095,	0x20000025,
                            0x20000009};
//   
static int _MathLib_Atanh[42] = {0,0x2327D4F5,	0x20B15DF5,	0x202B1239,	0x200AB116,
                        	0x2002AB11,	0x2000AAB1,	0x20002AAB,	0x20000AAB,
                            0x200002AB,	0x200000AB,	0x2000002B,	0x2000000B,
                            0x20000003,	0x20000001,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000,	0x20000000,	0x20000000,	0x20000000,
                            0x20000000};

#ifdef __cplusplus
extern "C" {
#endif

//       (   float)
inline unsigned int SrvExp2(unsigned int& Arg, int i, int& NewExp);

//      
//  (  Exp(x)   Exp(-x)   )
inline unsigned int SrvExp1(unsigned int Arg, int i);

// tanhf --
//
//     
//
//    --.
//    CORDIC,    
float tanhf(float f){

	int Mnt;                //   
    int Exp, Sgn, Spc;      //     
	int NewExp=0;           //      
    int PlusExp=0;          //      
	unsigned int Arg;       //   
    int sArg;               //     
    unsigned int rez;       //   
    unsigned int mines_rez; //    
    unsigned int iCosh;     //     
    unsigned int iSinh;     //     

    //   float
    DestructFloat( f, Mnt, Exp, Sgn, Spc );
    int y;                  //      

	//  
    if ( Exp ==128 && Mnt==0 )					//  4 NAN
		return f;

	if ( Exp > 6 ){			//     tanh(x) = 1
		return ConstructFloat( 0, 0, Sgn, 0);
	}

    if ( Exp < -12 )			//    
	    return f;		
    
	int j=-8;           //   
    //      ! 
	//     () ,  floor 
	if (Exp>=0){				
		Arg = Mnt << (Exp);
	}
	else if (Exp>=-2){
		Arg = Mnt << 4;
		Arg = Arg >> (-Exp);
		j=-4;
	}
	else {
		Arg = Mnt << 8;
		Arg = Arg >> (-Exp);
		j=0;
	}

    rez = SrvExp2( Arg, j, PlusExp);
    NewExp = PlusExp;
    iSinh = rez;
    iCosh = rez;
    if ( NewExp >15 ){              //      
	    if (NewExp>127){			//      tanh(x) = 1
			return ConstructFloat( 0, 0, Sgn, 0);	
	    }
    }
    else{                      
		Arg = TLN2 - Arg;
        mines_rez = SrvExp1(Arg, 0);
        iSinh -= mines_rez >> ((NewExp<<1)+1);
        iCosh += mines_rez >> ((NewExp<<1)+1);
    }

    NewExp=0;

    if(Exp<0){
        sArg = Mnt << 6;

        NewExp = Exp+1;

	    y = 0; 
        int x = _MathLib_iK[-NewExp];
	    int x1=x;

        int i = 1-NewExp;
        int k=i;

        const int iMax = 31;

        for (; i<iMax+NewExp;){
	        for (; i<k; i++){
		        if (sArg >= 0){
			        x1 += y >> (i-NewExp);
			        y += x >> (i+NewExp);
			        sArg = (sArg - _MathLib_Atanh[i])<<1;
		        }
		        else{
			        x1 -= y >> (i-NewExp);
			        y -= x >> (i+NewExp);
			        sArg = (sArg + _MathLib_Atanh[i])<<1;
		        }
		        x = x1;
	        }
            {
		        if (sArg >= 0){
			        x1 += y >> (i-NewExp);
			        y += x >> (i+NewExp);
			        sArg -= _MathLib_Atanh[i];
		        }
		        else{
			        x1 -= y >> (i-NewExp);
			        y -= x >> (i+NewExp);
			        sArg += _MathLib_Atanh[i];
		        }
		        x = x1;
            }
            k = (k<<1)+k+1;
            if (k>iMax+NewExp)
                k = iMax+NewExp;
	    }
	    while (i<30-NewExp){  //  
		    if (sArg >= 0){
			    y += x >> (i+NewExp);
			    sArg = (sArg - _MathLib_Atanh[i++])<<1;
		    }
		    else{
			    y -= x >> (i+NewExp);
			    sArg = (sArg + _MathLib_Atanh[i++])<<1;
		    }
		    x = x1;
	    }
        NewExp+=2-PlusExp;
        iSinh = y;
    }
    ULONG lSinh = iSinh;
    lSinh *= 0x40000000;
    y = lSinh / iCosh;
    if (y!=0){
		while (!(y & 0x40000000)){
			y <<= 1;
			NewExp--;
		}
	}
	else
		NewExp = -127;

	y >>=6;	//    , ...
	y = (y>>1) + (y & 1); //   ,  .

	if (y == 0x01000000){
		y = 0;
		NewExp++;
	}

    //   float
    return ConstructFloat( y, NewExp, Sgn, 0);	
}

//       (   float)
inline unsigned int SrvExp1(unsigned int Arg, int i){
    int tmp = Arg;
    unsigned int rez = 0x40000000;
	while (i!=30){  //  
		tmp -= _TLN[i++]; //    , ffrexp
        //    ,  TLN[]
        //  ,    
		if (tmp>0){
			rez += rez >> i; //  - !
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}
    return rez;
}

//      
//  (  Exp(x)   Exp(-x)   )
inline unsigned int SrvExp2(unsigned int& Arg, int i, int& NewExp){
    int tmp = Arg;
    unsigned int RetArg;
    unsigned int rez = 0x40000000;
	while (i<0){  //  
		i++;
		tmp -= TLN2; 
		if (tmp>0){
			Arg = tmp;
			NewExp += 1 << (-i);
		}
		else //    - 
			tmp = Arg;
		tmp <<=1;
		Arg = tmp;
	}
	tmp >>=1;
	Arg = tmp;
    RetArg = Arg;
	while (i!=30){  //  
		tmp -= _TLN[i++]; //    , ffrexp
                          //    ,  TLN[]
                          //  ,    
		if (tmp>0){
			rez += rez >> i; //  - !
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}
    Arg = RetArg;
    return rez;
}

#ifdef __cplusplus
}
#endif
